import 'package:codenic_exception_converter/src/failures/failure.dart';
import 'package:codenic_logger/codenic_logger.dart';
import 'package:fpdart/fpdart.dart';

/// {@template BaseExceptionConversion}
/// An abstract class that converts an exception [E] to failure [F] if exception
/// [E] occurs while running [call].
///
/// If no error occurs, then value [R] is returned.
/// {@endtemplate}
abstract class BaseExceptionConversion<R, E extends Exception,
    F extends Failure> {
  /// {@macro BaseExceptionConversion}
  const BaseExceptionConversion();

  /// Returns `true` if the [type] is the same as [E]. Otherwise, `false` is
  /// returned.
  bool checkExceptionType(Type type) {
    return type == E;
  }

  /// Executes the asynchronous [action].
  ///
  /// {@template call}
  /// If an [Exception] is thrown by the [action], then it will be converted
  /// into a [Failure] [F]. Otherwise, [R] will be returned.
  ///
  /// The [logger] and [messageLog] is used to log and give more details
  /// about the exception.
  /// {@endtemplate}
  Future<Either<Failure, R>> call({
    required Future<Either<Failure, R>> Function() action,
    CodenicLogger? logger,
    MessageLog? messageLog,
  }) async {
    assert(
      messageLog == null || logger != null,
      'If you want to log the exception, then you must provide a logger.',
    );

    try {
      return await action();
    } on E catch (exception, stackTrace) {
      return Left(
        onException(
          logger: logger,
          messageLog: messageLog,
          exception: exception,
          stackTrace: stackTrace,
        ),
      );
    }
  }

  /// Executes the synchronous [action].
  ///
  /// {@macro call}
  Either<Failure, R> callSync({
    required Either<Failure, R> Function() action,
    CodenicLogger? logger,
    MessageLog? messageLog,
  }) {
    assert(
      messageLog == null || logger != null,
      'If you want to log the exception, then you must provide a logger.',
    );

    try {
      return action();
    } on E catch (exception, stackTrace) {
      return Left(
        onException(
          logger: logger,
          messageLog: messageLog,
          exception: exception,
          stackTrace: stackTrace,
        ),
      );
    }
  }

  /// Converts the [Exception] [E] to a [Failure] [F].
  ///
  /// The [logger] and [messageLog] is used to log and give more details
  /// about the exception.
  F onException({
    required E exception,
    StackTrace? stackTrace,
    CodenicLogger? logger,
    MessageLog? messageLog,
  });
}
