import 'package:codenic_exception_converter/codenic_exception_converter.dart';
import 'package:codenic_exception_converter/src/exception_conversions/default_exception_conversion.dart';
import 'package:codenic_exception_converter/src/exception_conversions/socket_exception_conversion.dart';
import 'package:codenic_logger/codenic_logger.dart';
import 'package:fpdart/fpdart.dart';

/// A signature for [BaseExceptionConversion] factories.
typedef ExceptionConversionFactory
    = BaseExceptionConversion<T, Exception, Failure> Function<T>();

/// A suite of utility functions for converting [Exception]s into [Failure]s.
abstract class ExceptionConverter {
  const ExceptionConverter._();

  /// The global [BaseExceptionConversion]s used to convert [Exception]s into
  /// [Failure]s.
  ///
  /// This is used by [observe], [observeSync], and [convert].
  static List<ExceptionConversionFactory> exceptionConversions = [
    SocketExceptionConversion.new,
  ];

  /// The global logger used by [observe] and [observeSync].
  static CodenicLogger logger = CodenicLogger();

  static List<BaseExceptionConversion<T, Exception, Failure>>
      _exceptionConversions<T>() {
    return exceptionConversions.map((e) => e.call<T>()).toList();
  }

  /// Asynchronously converts any uncaught [Exception] thrown by the [action]
  /// into a [Failure].
  ///
  /// {@template observe}
  /// If no error occurs, then value [T] is returned and
  /// [logger.info(messageLog)] is called.
  ///
  /// If a [Failure] is returned by the [action], then [logger.warn()] is
  /// called.
  ///
  /// If an [Exception] is thrown, then it is handled by the
  /// [exceptionConversions] and [ExceptionConverter.exceptionConversions].
  ///
  /// If the [Exception] does not have an exception conversion, then it will be
  /// converted to [Failure] by default.
  ///
  /// NOTE:
  /// -  [Error]s are not caught by this method considering that it is bad
  /// practice to catch an [Error].
  /// See https://groups.google.com/a/dartlang.org/g/misc/c/lx9CXiV3o30/m/s5l_PwpHUGAJ.
  /// {@endtemplate}
  static Future<Either<Failure, T>> observe<T>({
    required MessageLog messageLog,
    required Future<Either<Failure, T>> Function(MessageLog messageLog) action,
    List<BaseExceptionConversion<T, Exception, Failure>>? exceptionConversions,
  }) async {
    final extendedExceptionConversions =
        _extendedExceptionConversions<T>(exceptionConversions);

    return extendedExceptionConversions
        .fold<Future<Either<Failure, T>> Function()>(
      () async {
        final result = await action(messageLog);

        result.fold(
          (l) => logger.warn(messageLog..message = '$l'),
          (r) => logger.info(messageLog..message = '$r'),
        );

        return result;
      },
      (previousValue, element) => () => element.call(
            action: previousValue,
            logger: logger,
            messageLog: messageLog,
          ),
    )();
  }

  /// Synchronously converts any uncaught [Exception] thrown by the [action]
  /// into a [Failure].
  ///
  /// {@macro observeSync}
  static Either<Failure, T> observeSync<T>({
    required Either<Failure, T> Function(MessageLog? messageLog) action,
    MessageLog? messageLog,
    List<BaseExceptionConversion<T, Exception, Failure>>? exceptionConversions,
  }) {
    final extendedExceptionConversions =
        _extendedExceptionConversions<T>(exceptionConversions);

    return extendedExceptionConversions.fold<Either<Failure, T> Function()>(
      () {
        final result = action(messageLog);

        if (messageLog != null) {
          result.fold(
            (l) => logger.warn(messageLog..message = 'Failed'),
            (r) => logger.info(messageLog..message = 'Success'),
          );
        }

        return result;
      },
      (previousValue, element) => () => element.callSync(
            action: previousValue,
            logger: logger,
            messageLog: messageLog,
          ),
    )();
  }

  /// Converts the given [error] into a [Failure].
  ///
  /// The [error] is expected to be an [Exception] object.
  ///
  /// If the given [error] is an [Error], then it is re-thrown. For more info,
  /// see https://stackoverflow.com/a/57004304.
  ///
  /// If the given [error] is not an [Exception], then an [ArgumentError] is
  /// thrown.
  static Failure convert({
    required MessageLog messageLog,
    required Object error,
    StackTrace? stackTrace,
    List<BaseExceptionConversion<void, Exception, Failure>>?
        exceptionConversions,
  }) {
    if (error is Error) {
      // Errors must not be caught.
      // See https://stackoverflow.com/a/57004304
      throw error;
    }

    if (error is! Exception) {
      throw ArgumentError.value(
        error,
        'error not an exception',
        'The error must be an Exception.',
      );
    }

    final extendedExceptionConversions =
        _extendedExceptionConversions<void>(exceptionConversions);

    for (final exceptionConversion in extendedExceptionConversions) {
      if (exceptionConversion.checkExceptionType(error.runtimeType)) {
        return exceptionConversion.onException(
          exception: error,
          messageLog: messageLog,
          logger: logger,
          stackTrace: stackTrace,
        );
      }
    }

    // coverage:ignore-start
    throw StateError(
      'No matching exception converesion found for ${error.runtimeType}',
    );

    // coverage:ignore-end
  }

  static List<BaseExceptionConversion<T, Exception, Failure>>
      _extendedExceptionConversions<T>(
    List<BaseExceptionConversion<T, Exception, Failure>>? exceptionConversions,
  ) =>
          [
            ...?exceptionConversions,
            ..._exceptionConversions<T>(),
            DefaultExceptionConversion<T>(),
          ];
}
