import 'dart:io';

import 'package:codenic_exception_converter/codenic_exception_converter.dart';
import 'package:codenic_exception_converter/src/exception_conversions/socket_exception_conversion.dart';
import 'package:codenic_exception_converter/src/failures/network_failure.dart';
import 'package:codenic_logger/codenic_logger.dart';
import 'package:fpdart/fpdart.dart';
import 'package:mocktail/mocktail.dart';
import 'package:test/test.dart';

class MockLogger extends Mock implements CodenicLogger {}

void main() {
  group(
    'Exception converter',
    () {
      final messageLog = MessageLog(id: 'test');
      final logger = MockLogger();
      ExceptionConverter.logger = logger;

      setUp(() {
        reset(logger);
      });

      group(
        'Observe sync',
        () {
          test(
            'should return a `NetworkFailure` when a socket exception is '
            'thrown',
            () {
              ExceptionConverter.observeSync<void>(
                messageLog: messageLog,
                exceptionConversions: [const SocketExceptionConversion()],
                action: (messageLog) {
                  throw const SocketException('test');
                },
              ).fold(
                (l) => expect(l, isA<NetworkFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a failure from the global exception conversions '
            'when an exception is thrown',
            () {
              ExceptionConverter.observeSync<void>(
                messageLog: messageLog,
                action: (messageLog) {
                  throw const SocketException('test');
                },
              ).fold(
                (l) => expect(l, isA<NetworkFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a base failure when an exception is thrown without '
            'an exception conversion',
            () {
              ExceptionConverter.observeSync<void>(
                messageLog: messageLog,
                action: (messageLog) {
                  throw const FormatException();
                },
              ).fold(
                (l) => expect(l, isA<Failure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test('should log a warning when a failure is returned, ', () {
            ExceptionConverter.observeSync<void>(
              messageLog: messageLog,
              action: (messageLog) {
                return const Left(Failure());
              },
            );

            verify(() => logger.warn(messageLog)).called(1);
          });

          test('should log an info on success', () {
            ExceptionConverter.observeSync<void>(
              messageLog: messageLog,
              action: (messageLog) {
                return const Right(null);
              },
            );

            verify(() => logger.info(messageLog)).called(1);
          });
        },
      );

      group(
        'Observe',
        () {
          test(
            'should return a `NetworkFailure` when a socket exception is '
            'thrown',
            () async {
              final result = await ExceptionConverter.observe<void>(
                messageLog: messageLog,
                exceptionConversions: [const SocketExceptionConversion()],
                action: (messageLog) {
                  throw const SocketException('test');
                },
              );

              result.fold(
                (l) => expect(l, isA<NetworkFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a failure from the global exception conversions '
            'when an exception is thrown',
            () async {
              final result = await ExceptionConverter.observe<void>(
                messageLog: messageLog,
                action: (messageLog) {
                  throw const SocketException('test');
                },
              );

              result.fold(
                (l) => expect(l, isA<NetworkFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a base failure when an exception is thrown without '
            'an exception conversion',
            () async {
              final result = await ExceptionConverter.observe<void>(
                messageLog: messageLog,
                action: (messageLog) {
                  throw const FormatException();
                },
              );

              result.fold(
                (l) => expect(l, isA<Failure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should log a warning when a failure is returned, ',
            () async {
              await ExceptionConverter.observe<void>(
                messageLog: messageLog,
                action: (messageLog) async {
                  return const Left(Failure());
                },
              );

              verify(() => logger.warn(messageLog)).called(1);
            },
          );

          test(
            'should log an info on success',
            () async {
              await ExceptionConverter.observe<void>(
                messageLog: messageLog,
                action: (messageLog) async {
                  return const Right(null);
                },
              );

              verify(() => logger.info(messageLog)).called(1);
            },
          );
        },
      );

      group(
        'convert',
        () {
          test(
            'should convert `SocketException to `NetworkFailure`',
            () {
              final result = ExceptionConverter.convert(
                messageLog: messageLog,
                error: const SocketException('test'),
                exceptionConversions: [const SocketExceptionConversion()],
              );

              expect(result, isA<NetworkFailure>());
            },
          );

          test(
            'should throw an `Error` when an `Error` is passed',
            () {
              expect(
                () => ExceptionConverter.convert(
                  messageLog: messageLog,
                  error: const OutOfMemoryError(),
                  exceptionConversions: [const SocketExceptionConversion()],
                ),
                throwsA(isA<OutOfMemoryError>()),
              );
            },
          );

          test(
            'should throw an `ArgumentError` when an a non-exception is passed',
            () {
              expect(
                () => ExceptionConverter.convert(
                  messageLog: messageLog,
                  error: 'Test',
                  exceptionConversions: [const SocketExceptionConversion()],
                ),
                throwsA(isA<ArgumentError>()),
              );
            },
          );
        },
      );
    },
  );
}
